(:~
 : This module serves for handling the semantic splitting of the complete XML files.
 : For displaying the XML data in HTML we first have to assemble all XML that
 : belong to a BdN work and then split it into smaller chunks (else each web page
 : is too big to display it in a user-friendly way).
 :
 : The splitting is handled according to the structure given at bdn:body-structure
 : in whitelist.xml. This approach has been chosen due to the structural variety
 : of the BdN works.
 :
 : @author Michelle Weidling
 :)

xquery version "3.1";

module namespace sem-split="http://bdn-edition.de/portal/sem-split";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace bdn="http://www.bdn-edition.de/ns/1.0";
declare namespace util="http://exist-db.org/xquery/util";

import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace functx="http://www.functx.com";
import module namespace ifweb="http://bdn-edition.de/if/ifweb" at "../if/ifweb.xqm";
import module namespace whitelist="https://bdn-edition.de/ns/whitelist" at "whitelist.xqm";


declare option exist:serialize "method=xml media-type=text/xml expand-xincludes=no";

(:~
 : Splits a BdN work into smaller chunks according to a given BdN author. The
 : current splitting results in separate XML files for tei:group/tei:text/tei:front,
 : tei:group/tei:text/tei:back, tei:group/tei:text//tei:div[@type = 'chapter']
 : and everything before a tei:div[@type = 'part' or @type = 'chapter']
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @param $file The full XML file (in memory)
 : @return the document path to the stored XML files
 :)
declare function sem-split:main($author as xs:string,
$file as node()*) as xs:string+
{
    util:log("info", "### START SEMANTIC SPLITTING ###"),
    sem-split:to-foreword($author, $file),
    sem-split:to-volumes($author, $file)
};


(:~
 : Creates the modern edition's foreword XML for a given BdN author.
 :
 : @author Michelle Weidling
 : @param $author the BdN author's name
 : @param $document The full XML file
 : @return The location where the XML is stored.
 :)
declare function sem-split:to-foreword($author as xs:string,
$document as node()*) as xs:string+
{
    util:log("info", "Start split to foreword."),
    xmldb:store(
        bdn-utils:get-xml-dir($author),
        $author || "-foreword.xml",
        sem-split:prepare-foreword-xml($document//tei:front[not(ancestor::tei:group)])
    )
};


(:~
 : An auxiliary function that adds IDs to all tei:note elements in a modern
 : editor's foreword.
 :
 : @author Michelle Weidling
 : @param $nodes The nodes of the foreword
 : @return The processed nodes
 :)
declare function sem-split:prepare-foreword-xml($nodes as node()*) as node()*
{
    for $node in $nodes return
        typeswitch ($node)
        case text() return
            $node

         case comment() return
             ()

        case element(tei:note) return
            element tei:note
            {
                attribute id {generate-id($node)},
                $node/@*,
                sem-split:prepare-foreword-xml($node/node())
            }

        default return
            element {QName("http://www.tei-c.org/ns/1.0", $node/name())}
            {
                $node/@*,
                sem-split:prepare-foreword-xml($node/node())
            }
};


(:~
 : Splits a BdN work into its volumes.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @param $document $author || "-full.xml" (in memory)
 : @return The document path of the created XML files
 :)
declare function sem-split:to-volumes($author as xs:string,
$document as node()*) as xs:string+
{
    let $add-log-entry := util:log("info", "Start splitting to volumes.")
    let $volumes := $document//tei:group/tei:text
    let $vol-no := count($volumes)
    let $body-structure := whitelist:get-doc()//bdn:author[./string() = $author]/following-sibling::bdn:body-structure
    let $bla := util:log("info", "No of volumes: " || $vol-no)
    let $bla := util:log("info", "Body structure: " || serialize($body-structure))
    let $dir := bdn-utils:get-xml-dir($author)

    for $iii in 1 to $vol-no return
        (
            util:log("info", "In volume no " || $iii || ". " || "Corresponding body-structure entry: "
              || serialize($body-structure/bdn:volume[@n = xs:string($iii)]) || "."), (: sack und steinbart sind hier empty!:)
            sem-split:to-front-body-back($author, $iii, $volumes[$iii], $body-structure/bdn:volume[@n = xs:string($iii)])
        )
};



(:~
 : Splits each volume of a BdN work into its respective front, body and back part.
 : While front and back are stored at this point, the tei:body is treated
 : differently according to the current author and his requirements.
 :
 : Currently, only Teller and Leß have special requirements for splitting tei:body
 : any further.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $author The current author
 : @param $vol-no The number of the current volume in the BdN work
 : @param $volume The current volume
 : @param $whitelist-entry The bdn:volume element in whitelist.xml that refers to the current volume
 : @return The document path of the created XML files
 :)
declare function sem-split:to-front-body-back($author as xs:string, $vol-no
as xs:integer, $volume as node(), $whitelist-entry as element(bdn:volume))
as xs:string*
{
    let $volume-children := $volume/child::*
    let $children-no := count($volume-children)
    let $file-prefix := $author || "-" || $vol-no
    let $dir := bdn-utils:get-xml-dir($author)

    for $iii in 1 to $children-no
        let $content := $volume-children[$iii]
        let $add-log-entry := util:log("info", "Splitting to a volume's front, body, and back. Currently in child " || name($content) || ".")
        return
            if ($content[self::tei:front]) then
                let $add-log-entry := util:log("info", "Split to a volume's front.")
                let $content := sem-split:make-split-div($content, false())
                return
                  (
                    sem-split:save($dir, $file-prefix, "front", $content),
                    util:log("info", "Front XML created.")
                  )
            else if ($content[self::tei:back]) then
                let $add-log-entry := util:log("info", "Split to a volume's back.")
                let $content := sem-split:make-split-div($content, false())
                return
                  (sem-split:save($dir, $file-prefix, "back", $content),
                  util:log("info", "Back XML created."))

            (: author specific treatment :)
            else if ($author = "teller") then
                sem-split:to-letters($file-prefix, $content, $dir)
            else if ($author = "less") then
                sem-split:to-sermons($file-prefix, $content, $dir)

            else
                sem-split:body-to-substructure($author || "-" || $vol-no, $dir, $author, $vol-no, $whitelist-entry)
};


(:~ An auxiliary function that puts any given nodes in a tei:div container. By
 : @class = 'split' the division is marked as being created during
 : the splitting process.
 :
 : @author Michelle Weidling
 : @param $content Some TEI nodes
 : @param $target true() if the corresponding node in bdn:body-structure has a @target="body", e.g. in Bahrdt-Semler
 : @see https://gitlab.gwdg.de/bibliothek-der-neologie/bdn-SADE/issues/147
 : @return A tei:div containing $content
 :)
declare function sem-split:make-split-div($content as node()+, $target as xs:boolean)
as element(tei:div)
{
  element tei:div
  {
    attribute class {"split"},
    if ($target) then
      attribute xml:id {$content/ancestor-or-self::tei:text[1]/@xml:id}
    else
      (),
    $content
  }
};



(:~ An auxiliary function for saving files as standard and intermediate format.
 :
 : @author Michelle Weidling
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1"
 : @param $file-appendix e.g. "front" or "back"
 : @param $nodes The content to be saved
 : @return Two strings indicating the path to the saved files
 :)
declare function sem-split:save($dir as xs:string, $file-prefix as xs:string, $file-appendix, $content as node()+)
as xs:string+
{
  xmldb:store($dir, $file-prefix || "-" || $file-appendix || "-NIF.xml", $content),
  ifweb:transform-split-xml($content, $file-prefix || "-" || $file-appendix || ".xml", $dir)
};


(:~
 : Splits a tei:body element into further chunks if necessary. Whether this is
 : done or if the tei:body is saved as a whole is determined by bdn:body-structure
 : given in whitelist.xml: If a bdn:volume has any children, then the current
 : tei:body is split up further. These children correspond to the entries of the
 : the printed TOC.
 :
 : @author Michelle Weidling
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1". This is used for storing the resulting XML files.
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @param $author The current author, e.g. "noesselt"
 : @param $volume-no The current volume's number, e.g. "1"
 : @param $whitelist-entry The bdn:volume that correspongs to the current volume
 : @return A sequence of strings containing the paths to the storedfiles
 :)
declare function sem-split:body-to-substructure($file-prefix as xs:string,
$dir as xs:string, $author as xs:string,
$volume-no as xs:string, $whitelist-entry as element(bdn:volume)) as xs:string+
{
    let $add-log-entry := util:log("info", "Start splitting the body structure.\n -- Start splitting the body structure." ||
     "-- File prefix: " || $file-prefix || "-- Volume-no: " || $volume-no || "-- Whitelist entry: " || serialize($whitelist-entry))

    let $full-file := doc($config:full-xmls || $author || "-full.xml")/*

    let $front := $full-file//tei:group/tei:text[xs:integer($volume-no)]/tei:front
    let $back := $full-file//tei:group/tei:text[xs:integer($volume-no)]/tei:back

    return
        (
            (: PROBLEM: relies on the existence on front instead of scope info only; in B/S front and body may be merged sometimes:)
            if ($front) then
                sem-split:to-front($file-prefix, $front, $dir)
            else
                (),

            (: in this case a volume doesn't have any other underlying structure :)
            if ($whitelist-entry[not(child::*) and @scope]) then
                (
                    util:log("info", "NAY: This volume doesn't have any underlying structure."),
                    sem-split:save-complete-body-element($file-prefix, $dir, $full-file, $volume-no)
                )
            (: in this case a volume has further structure, e.g. chapters or prefaces :)
            else
                (
                    util:log("info", "YAY: This volume has any underlying structure."),
                    sem-split:split-body($file-prefix, $whitelist-entry//*[@scope and not(child::*)], $dir, $full-file)
                ),

            if ($back) then
                sem-split:to-back($file-prefix, $back, $dir)
            else
                ()
        )
};


(:~
 : Creates a wrapper tei:div and stores the front part of a volume in a separate
 : file. Since we need the intermediate format version of it to display the
 : scribal abbreviations correctly, this one is created, too.
 :
 : @author Michelle Weidling
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1". This is used for storing the resulting XML files.
 : @param $front The current volume's tei:front node
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @return The location of the stored files
 :)
declare function sem-split:to-front($file-prefix as xs:string, $front as node()*,
$dir as xs:string) as xs:string+
{
    sem-split:save($dir, $file-prefix, "front", sem-split:make-split-div($front, false())),
    util:log("info", "Front XML created.")

};


(:~
 : In some cases a given volume doesn't have underlying structure elements like
 : chapters or dedications. In this case we simply save its whole tei:body
 : element since we don't have any criteria for determining where to split it.
 :
 : As a consequence, the navigation of the web page always has the same structure
 : as the print edition.
 :
 : @author Michelle Weidling
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1". This is used for storing the resulting XML files.
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @param $full-file The tei:TEI node of the complete work XML (e.g. noesselt-full.xml)
 : @param $volume-no  The current volume's number, e.g. "1"
 : @return The location of the stored files
 :)
declare function sem-split:save-complete-body-element($file-prefix as xs:string,
$dir as xs:string, $full-file as node()*, $volume-no as xs:string)
as xs:string+
{
    let $add-log-entry := util:log("info", "-- Saving complete body element.")
    let $body := $full-file//tei:group/tei:text[xs:integer($volume-no)]/tei:body
    let $volume-id :=
        if ($body/@xml:id) then
            ()
        else
            $body/ancestor::tei:text[1]/@xml:id
    let $body-with-id :=
        if ($body/@xml:id) then
            ()
        else
            functx:add-attributes($body, xs:QName("xml:id"), $volume-id)
    let $content := sem-split:make-split-div($body-with-id, false())

    return
    (
        sem-split:save($dir, $file-prefix, $volume-id, $content),
        util:log("info", "Body XML created without any further splitting.")
    )
};



(:~
 : Splits Teller's dictionary along the letters.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file-prefix The current file-prefix, e.g. "noesselt-1-1"
 : @param $document The tei:body of the edition
 : @param $dir The directory where to save the result XML files
 : @return The document path of the created XML files
 :)
declare function sem-split:to-letters($file-prefix as xs:string, $document as
node()*, $dir as xs:string) as xs:string+
{
    let $letters := $document//tei:div[@type = "letter"]
    let $letter-no := count($letters)

    for $iii in 1 to $letter-no
        let $current-letter := functx:substring-after-last($letters[$iii]/@xml:id, "_")
        let $filename := $file-prefix || $current-letter || ".xml"
        return
            (
                xmldb:store($dir, $file-prefix || $current-letter || "-NIF.xml", $letters[$iii]),
                ifweb:transform-split-xml($letters[$iii], $filename, $dir),
                util:log("info", "Stored Teller's lemma " || $current-letter)
            )
};

(:~
 : Splits Less' sermon anthology into it's sections (= sermons).
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file-prefix The current file-prefix, e.g. "noesselt-1-1"
 : @param $document The tei:body of the edition
 : @param $dir The directory where to save the result XML files
 : @return The document path of the created XML files
 :)
declare function sem-split:to-sermons($file-prefix as xs:string, $document as
node()*, $dir as xs:string) as xs:string+
{
    let $add-log-entry := util:log("info", "Start semantic splitting to sermons.")

    let $sermons := $document/child::*
    let $sermons-no := count($sermons)

    for $iii in 1 to $sermons-no
        let $filename := $file-prefix || "-" || $iii || ".xml"

        return
            (xmldb:store($dir, $file-prefix || "-" || $iii || "-NIF.xml", $sermons[$iii]),
            ifweb:transform-split-xml($sermons[$iii], $filename, $dir),
            util:log("info", "Stored Less' sermon " || $iii))
};


(:~
 : When a volume has further underlying structure, this is marked by it having
 : child elements in bdn:body-structure. The naming of these child elements
 : depend on the respective element that should appear in the TOC/navigation.
 : For each child element a separate XML file is created, combined with pre-unit
 : nodes (if stated by @pre-unit = 'true' in a child element).
 :
 : As a consequence, the navigation of the web page always has the same structure
 : as the print edition.
 :
 : @author Michelle Weidling
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1". This is used for storing the resulting XML files.
 : @param $units The child elements of a given volume in bdn:body-structure, e.g. bdn:letter or bdn:chapter
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @param $full-file The tei:TEI node of the complete work XML (e.g. noesselt-full.xml)
 : @return The location of the stored files
 :)
declare function sem-split:split-body($file-prefix as xs:string, $units as node()*,
$dir as xs:string, $full-file as node()*) as xs:string+
{
    util:log("info", "---- Start further splitting of body. Involved units: " ||  serialize($units)),
    for $unit in $units return
        let $scopes := tokenize($unit/@scope, " ")
        let $scope-ids :=
            for $scope in $scopes return
                replace($scope, "#", "")
        return
              try
              {
                  let $elements-referred-to :=
                      for $id in $scope-ids return
                          if ($unit/@pre-unit = "true" and matches($unit/@pre-unit-ref, $id)) then
                              ($full-file//*[@xml:id = $id]/preceding-sibling::*, $full-file//*[@xml:id = $id])
                          else
                              $full-file//*[@xml:id = $id]

                  let $elements-referred-to :=
                    if ($unit/@target = "body") then
                        let $add-log-entry := util:log("info", "---- unit with @target: " || serialize($elements-referred-to))
                        return
                            $elements-referred-to//tei:body

                    else
                        $elements-referred-to

                  let $add-log-entry :=
                      if ($elements-referred-to) then
                          util:log("info", "---- Elements referred to found!")
                      else
                          util:log("info", "---- Elements referred to NOT FOUND...")

                  let $file-prefix := $file-prefix || "-" || string-join($scope-ids, "-")
                  let $add-log-entry := util:log("info", "---- file prefix: " || serialize($file-prefix))
                  let $content :=
                    if($unit/@target = "body") then
                      sem-split:make-split-div($elements-referred-to, true())
                    else
                        sem-split:make-split-div($elements-referred-to, false())
                  (: let $add-log-entry := util:log("debug", "---- content: " || serialize($content)) :)

                  return
                    (
                        xmldb:store($dir, $file-prefix || "-NIF.xml", $content),
                        ifweb:transform-split-xml($content, $file-prefix || ".xml", $dir),
                        util:log("info", "New file " || $file-prefix || " saved based on " || serialize($unit))
                    )
                }
                catch *
                {
                    util:log("error", "Seems like there are some IDs in @scope wrong."),
                    error(QName("error", "SEMSPLIT001"), "Seems like there are some IDs in @scope wrong.")
                }
};


(:~
 : Creates a wrapper tei:div and stores the back part of a volume in a separate
 : file. Since we need the intermediate format version of it to display the
 : scribal abbreviations correctly, this one is created, too.
 :
 : @author Michelle Weidling
 : @param $file-prefix The current file prefix of the volume, e.g. "noesselt-1". This is used for storing the resulting XML files.
 : @param $front The current volume's tei:back node
 : @param $dir The path of the collection to store the resulting XML files to, e.g. "/db/apps/sade/resources/edition/noesselt/xml/"
 : @return The location of the stored files
 :)
declare function sem-split:to-back($file-prefix as xs:string, $back as node()*,
$dir as xs:string) as xs:string+
{
    sem-split:save($dir, $file-prefix, "back", sem-split:make-split-div($back, false())),
    util:log("info", "Back XML created.")
};
